local _G = getfenv(0)
local ADDON_NAME, addon = ...

-- Local versions for performance
local string = _G.string
local math = _G.math
local pairs = _G.pairs
local ipairs = _G.ipairs
local table = _G.table
local floor, ceil, min, max, abs = math.floor, math.ceil, math.min, math.max, math.abs
local tconcat, tinsert = table.concat, table.insert
local wipe = _G.wipe
local tostring = _G.tostring
local tonumber = _G.tonumber

local DKDiseases = _G.LibStub("AceAddon-3.0"):NewAddon(addon.addonName, "AceConsole-3.0", "AceEvent-3.0","AceTimer-3.0")
local L = _G.LibStub("AceLocale-3.0"):GetLocale(addon.addonName, true)
local LDB = _G.LibStub("LibDataBroker-1.1")
local LibQTip = _G.LibStub("LibQTip-1.0")
local Icon = _G.LibStub("LibDBIcon-1.0")
local LSM = _G.LibStub:GetLibrary("LibSharedMedia-3.0")

addon.options = nil
local MIN_UPDATE_TIME = addon.MIN_UPDATE_TIME
local round = addon.round
local FormatNumber = addon.FormatNumber

local HEXFMT = "|c%2x%2x%2x%2x"
local function chatColorCode(color)
	local a = 255
	local r = ceil(255 * color.r)
	local g = ceil(255 * color.g)
	local b = ceil(255 * color.b)
	return HEXFMT:format(a, r, g, b)
end

DKDiseases.registeredDisplays = {}
DKDiseases.displays = {
	target = nil,
	focus = nil,
}
DKDiseases.diseases = {
	target = {
		ff = nil,
		bp = nil,
		np = nil,
	},
	focus = {
		ff = nil,
		bp = nil,
		np = nil,
	},
}

addon.units = {"target", "focus"}
addon.current = {
	target = nil,
	focus = nil,
}
addon.hostile = {
	target = nil,
	focus = nil,
}
addon.isDK = false
addon.currentSpec = ""
addon.effectiveAP = 0
addon.mastery = 0
addon.critChance = 0
addon.spellCrit = 0
addon.damageModifier = 1.0
addon.frostVulnerability = 1.0
addon.avgWeaponDmg = 0
-- State for Soul Reaper Tracking
addon.soulReaper1 = 0
addon.soulReaper2 = 0
addon.howlingBlast = 0
addon.soulReaperMin = 55336
addon.soulReaperMax = 64309
addon.soulReaperBase = (addon.soulReaperMin + addon.soulReaperMax) / 2
addon.soulReaperRange = nil

local Broker = _G.CreateFrame("Frame")
Broker.obj = LDB:NewDataObject(_G.GetAddOnMetadata(ADDON_NAME, "Title"), {
	type = "data source",
	icon = "Interface\\Icons\\Spell_DeathKnight_BloodPlague",
	label = _G.GetAddOnMetadata(ADDON_NAME, "Title"),
	text = _G.GetAddOnMetadata(ADDON_NAME, "Title"),
	barValue = 0,
	barR = 0,
	barG = 0,
	barB = 1,
	OnClick = function(clickedframe, button)
		if button == "RightButton" then
			local optionsFrame = _G.InterfaceOptionsFrame

			if optionsFrame:IsVisible() then
				optionsFrame:Hide()
			else
				DKDiseases:ShowOptions()
			end
		end
	end
})

local GREEN = "|cff00ff00"
local YELLOW = "|cffffff00"
local BLUE = "|cff0198e1"
local ORANGE = "|cffff9933"
local FF_COLOR = chatColorCode({r = 0.0, g = 0.45*1.6, b = 0.6*1.6})
local BP_COLOR = chatColorCode({r = 0.5*1.6, g = 0.0, b = 0.6*1.6})
local Headers = {
	Addon = GREEN.."%s %s",
}

function Broker.obj:OnEnter()
	local tooltip = LibQTip:Acquire("DKDiseasesTooltip", 2, "LEFT", "RIGHT")
	self.tooltip = tooltip 

	tooltip:AddHeader(Headers.Addon:format(addon.addonTitle, addon.addonVersion))
	tooltip:AddLine()

	if addon.isDK then
	end

	tooltip:SmartAnchorTo(self)
	tooltip:Show()
end

function Broker.obj:OnLeave()
	LibQTip:Release(self.tooltip)
	self.tooltip = nil
end

local LDBDataFeed = false
local DataFeed = {
	display = "",
}

local function UpdateLDBData()
	Broker.obj.text = _G.GetAddOnMetadata(ADDON_NAME, "Title")
end

function DKDiseases:SetLDBDataDisplay()
	DataFeed.display = self.db.profile.LDBDataDisplay
	LDBDataFeed = DataFeed.display == "None" and false or true
	UpdateLDBData()
end

local LookupOrKeyMT = {__index = function (t,k) return k end}
local SpellIds = {
	["Frost Fever"] = 55095,
	["Blood Plague"] = 55078,
	["Necrotic Plague"] = 155163,
	["Plague Strike"] = 45462,
	["Outbreak"] = 77575,
	["Auto Attack"] = 6603,
	["Frost Vulnerability"] = 51714,
	["Howling Blast"] = 49184,
	["Soul Reaper"] = 130735,
	["Soul Reaper (Blood)"] = 114866,
	["Soul Reaper (Frost)"] = 130735,
	["Soul Reaper (Unholy)"] = 130736,
}
addon.SpellIds = SpellIds
local SpellNames = {}
local SpellIcons = {}
_G.setmetatable(SpellNames, LookupOrKeyMT)
local function LoadSpellNames()
	for k, v in pairs(SpellIds) do
		local name, _, icon = _G.GetSpellInfo(v)
		SpellNames[k] = name
		SpellIcons[k] = icon
	end
end
LoadSpellNames()
addon.SpellIcons = SpellIcons
addon.SpellNames = SpellNames

addon.soulReaperIds = {
	["Blood"] = SpellIds["Soul Reaper (Blood)"],
	["Frost"] = SpellIds["Soul Reaper (Frost)"],
	["Unholy"] = SpellIds["Soul Reaper (Unholy)"],
}

local frames = {}

local defaults = {
	profile = {
		minimap = {
			hide = true,
		},
		LDBDataDisplay = "None",
		locked = false,
		debug = false,
		specs = {
			["Frost"] = true,
			["Unholy"] = true,
			["Blood"] = true,
		},
		font_face = "Friz Quadrata TT",
		font_size = 13,
		fontFlags = {
			OUTLINE = true,
			THICKOUTLINE = false,
			MONOCHROME = false
		},
		texture = "Blizzard",
		diseaseOrder = "BF",
		units = {
			target = {
				enabled = true,
				hideooc = false,
				ooc_alpha = 0.5,
				combatAlpha = 1.0,
				hideNotAttackable = false,
				hideNoDiseases = false,
				hideInVehicle = true,
			},
			focus = {
				enabled = false,
				hideooc = false,
				ooc_alpha = 0.5,
				combatAlpha = 1.0,
				hideNotAttackable = false,
				hideNoDiseases = false,
				hideInVehicle = true,
			},
		},
		srFrame = {
			enabled = false,
			userPlaced = false,
			x = 0,
			y = 0,
			height = 32,
			width = 32,
			showText = true,
			percentOverride = false,
			percent = 0.35,
			alwaysShow = false,
			onlyShowHB = false,
			ignoreRuneCD = true,
			alphas = {
				usable = 1.0,
				cooldown = 0.5,
				unusable = 0.0,
				ooc = 0.0,
			},
			specs = {
				["Frost"] = true,
				["Unholy"] = false,
				["Blood"] = false,
			},
		},
		display = {
			target = "Block",
			focus = "Block",
		},
		displays = {
			["Block"] = {
				['**'] = {
					bkgcolor = { r = 0.0, g = 0.0, b = 0.5, a = 0.8 },
					textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
					timercolor = { r = 1.0, g = 1.0, b = 0.0, a = 1.0 },
					stackscolor = { r = 1.0, g = 1.0, b = 0.0, a = 1.0 },
					timer = true,
					timerFontSize = 14,
					stacks = true,
					stacksFontSize = 14,
					cooldown = true,
					reverseCooldown = false,
					background = true,
					x = 0,
					y = 0,
					attached = true,
					inset = 3,
					width = 35,
					height = 35,
					["Blood Plague"] = {
						x = -20,
						y = 0,
					},
					["Frost Fever"] = {
						x = 20,
						y = 0,
					},
					["Necrotic Plague"] = {
						x = 0,
						y = 0,
					},
				},
			},
			["Bar"] = {
				['**'] = {
					barcolor = { r = 0.3, g = 0.3, b = 1.0, a = 0.8 },
					bkgcolor = { r = 0.0, g = 0.0, b = 0.4, a = 0.8 },
					textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
					timercolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
					stackscolor = { r = 1.0, g = 1.0, b = 0.0, a = 1.0 },
					x = 0,
					y = 0,
					attached = true,
					width = 100,
					height = 20,
					value = true,
					timer = true,
					iconEnabled = true,
					iconPos = "LEFT",
					timerValueEnabled = true,
					timerFontSize = 13,
					timerPos = "RIGHT",
					stacks = true,
					stacksFontSize = 13,
					stacksPos = "LEFT",
					["Blood Plague"] = {
						x = 0,
						y = 10,
					},
					["Frost Fever"] = {
						x = 0,
						y = -10,
					},
					["Necrotic Plague"] = {
						x = 0,
						y = 0,
					},
				},
			},
		},
	},
}

function DKDiseases:GetOptions()
	if not addon.options then
		addon.options = {
			type = "group",
			name = addon.addonTitle,
			args = {
				core = self:GetCoreOptions(),
			}
		}
		addon.options.args.profile = 
			_G.LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)

	end
	return addon.options
end

function DKDiseases:ShowOptions()
	_G.InterfaceOptionsFrame_OpenToCategory(self.optionsFrame.Main)
end

function DKDiseases:GetCoreOptions()
	local core = {
		order = 1,
		name = L["General Options"],
		type = "group",
		inline = true,
		args = {
			description = {
				order = 1,
				type = "description",
				name = L["Addon_Desc"],
			},
			generalOptions = {
				order = 100,
				type = "group",
				name = L["General Options"],
				args = self:GetGeneralOptions(),
			},
			fonts = {
				order = 600,
				type = "group",
				name = L["Font"],
				args = {
					font_size = {
						order = 610,
						name = L["Font size"],
						desc = L["Font size"],
						type = "range",
						min = 8,
						max = 30,
						step = 1,
						set = function(info, val) 
							self.db.profile.font_size = val
							self:ResetFonts()
						end,
						get = function(info,val) return self.db.profile.font_size end,
					},
					font_face = {
						order = 620,
						type = "select",
						name = L["Font"],
						desc = L["Font to use."],
						values = LSM:HashTable("font"),
						dialogControl = 'LSM30_Font',
						get = function() return self.db.profile.font_face end,
						set = function(info, val) 
							self.db.profile.font_face = val
							self:ResetFonts()
						end
					},
					font_outline = {
						name = L["Outline"],
						desc = L["FontOutlineDesc"],
						type = "toggle",
						order = 630,
						set = function(info, val)
							self.db.profile.fontFlags.OUTLINE = val
							self:ResetFonts()
						end,
						get = function(info)
							return self.db.profile.fontFlags.OUTLINE
						end,
					},
					font_thickoutline = {
						name = L["Thick Outline"],
						desc = L["FontThickOutline_OptionDesc"],
						type = "toggle",
						order = 640,
						set = function(info, val)
							self.db.profile.fontFlags.THICKOUTLINE = val
							self:ResetFonts()
						end,
						get = function(info)
							return self.db.profile.fontFlags.THICKOUTLINE
						end,
					},
					font_monochrome = {
						name = L["Monochrome"],
						desc = L["FontMonochromeDesc"],
						type = "toggle",
						order = 650,
						set = function(info, val)
							self.db.profile.fontFlags.MONOCHROME = val
							self:ResetFonts()
						end,
						get = function(info)
							return self.db.profile.fontFlags.MONOCHROME
						end,
					},
				},
			},
			textures = {
				order = 700,
				type = "group",
				name = L["Textures"],
				args = {
					bartexture = {
						order = 110,
						name = L["Bar Texture"],
						desc = L["BarTexture_OptionDesc"],
						type = "select",
						values = LSM:HashTable("statusbar"),
						dialogControl = 'LSM30_Statusbar',
						get = function()
							return self.db.profile.texture
						end,
						set = function(info, val)
							self.db.profile.texture = val
							self:UpdateTextures()
						end,
					},
				},
			},
		},
	}

	-- Add additional options for modules
	local modnum = 1
	for name, obj in pairs(addon.modules) do
		if obj and obj.GetOptions then
			local modname, title, opts = obj:GetOptions()
			core.args[modname] = {
				order = 1000 * modnum,
				type = "group",
				name = title,
				args = opts
			}
			modnum = modnum + 1
		end
	end

	for i, unit in ipairs(addon.units) do
		local i = 100

		core.args[unit] = {
	        order = (unit == "target" and 300 or 400),
	        type = "group",
	        name = (unit == "target" and L["Target Frame"] or L["Focus Frame"]),
			args = {
				enabled = {
					order = 1,
					name = L["Enabled"],
					desc = L["Enabled"],
					type = "toggle",
					set = function(info, val)
						self.db.profile.units[unit].enabled = val
						if val then
							self:CreateOrUpdateDisplay(unit)
						elseif self.displays[unit] and self.displays[unit].Disable then
							self.displays[unit]:Disable()
						end
					end,
					get = function(info)
						return self.db.profile.units[unit].enabled
					end,
				},
				display = {
					order = 2,
					name = L["Display"],
					desc = L["Display"],
					type = "select",
					values = {
					    ["Bar"] = L["Bar"],
					    ["Block"] = L["Block"],
					},
					set = function(info, val)
						self.db.profile.display[unit] = val
						self:CreateOrUpdateDisplay(unit)
					end,
					get = function(info)
						return self.db.profile.display[unit]
					end,
				},
				ooc_alpha = {
					order = 3,
					name = L["OOC Alpha"],
					desc = L["OOCAlpha_Desc"],
					type = "range",
					min = 0,
					max = 1,
					step = 0.05,
					set = function(info, val)
					    self.db.profile.units[unit].ooc_alpha = val
						if self.displays[unit] then
							self.displays[unit]:UpdateLayout()
						end
					end,
					get = function(info, val)
					    return self.db.profile.units[unit].ooc_alpha
					end,					
				},
				combatAlpha = {
					order = 4,
					name = L["Combat Alpha"],
					desc = L["CombatAlpha_Desc"],
					type = "range",
					min = 0,
					max = 1,
					step = 0.05,
					set = function(info, val)
					    self.db.profile.units[unit].combatAlpha = val
						if self.displays[unit] then
							self.displays[unit]:UpdateLayout()
						end
					end,
					get = function(info, val)
					    return self.db.profile.units[unit].combatAlpha
					end,					
				},
				hideWhen = {
			        order = 100,
			        type = "header",
			        name = L["Hide When"],
				},
				hideooc = {
					order = 101,
					name = L["Out of Combat"],
					desc = L["Out of Combat"],
					type = "toggle",
					set = function(info, val)
					    self.db.profile.units[unit].hideooc = val
						if self.displays[unit] then
							self.displays[unit]:Show()
						end
					end,
		            get = function(info)
						return self.db.profile.units[unit].hideooc
					end,
				},
				hideNotAttackable = {
					order = 102,
					name = L["Not Attackable"],
					desc = L["Not Attackable"],
					type = "toggle",
					set = function(info, val)
					    self.db.profile.units[unit].hideNotAttackable = val
						if self.displays[unit] then
							self.displays[unit]:Show()
						end
					end,
		            get = function(info)
						return self.db.profile.units[unit].hideNotAttackable
					end,
				},
				hideNoDiseases = {
					order = 103,
					name = L["No Diseases"],
					desc = L["No Diseases"],
					type = "toggle",
					set = function(info, val)
					    self.db.profile.units[unit].hideNoDiseases = val
						if self.displays[unit] then
							self.displays[unit]:Show()
						end
					end,
		            get = function(info)
						return self.db.profile.units[unit].hideNoDiseases
					end,
				},
				hideInVehicle = {
					order = 103,
					name = L["In Vehicle"],
					desc = L["In Vehicle"],
					type = "toggle",
					set = function(info, val)
					    self.db.profile.units[unit].hideInVehicle = val
						if self.displays[unit] then
							self.displays[unit]:Show()
						end
					end,
		            get = function(info)
						return self.db.profile.units[unit].hideInVehicle
					end,
				},
			},
		}

		for name, display in pairs(self.registeredDisplays) do
			if display and display.GetOptions then
				local opts = display:GetOptions(unit)
				if opts then
					core.args[unit].args[name] = {
						order = i,
						name = name,
						type = "group",
						args = opts,
					}
					i = i + 1
				end
			end
		end
	end

	return core
end

function DKDiseases:GetGeneralOptions()
	local options = {
		locked = {
			name = L["Lock"],
			desc = L["LockDesc"],
			type = "toggle",
			order = 10,
			set = function(info, val)
				self.db.profile.locked = val
				for name, display in pairs(self.registeredDisplays) do
					if display and display.Lock then
						display:Lock(val)
					end
				end
				if addon.srFrame then addon.srFrame:Lock(val) end
			end,
			get = function(info)
				return self.db.profile.locked
			end,
		},
		minimap = {
			order = 20,
			name = L["Minimap Button"],
			desc = L["Toggle the minimap button"],
			type = "toggle",
			set = function(info,val)
				-- Reverse the value since the stored value is to hide it
				self.db.profile.minimap.hide = not val
				if self.db.profile.minimap.hide then
					Icon:Hide("DKDiseasesLDB")
				else
					Icon:Show("DKDiseasesLDB")
				end
			end,
			get = function(info)
				-- Reverse the value since the stored value is to hide it
				return not self.db.profile.minimap.hide
			end,
		},
		diseaseOrder = {
			order = 30,
			name = L["Disease Order"],
			desc = L["DiseaseOrder_Desc"],
			type = "select",
			values = {
				["FB"] = SpellNames["Frost Fever"]..", "..SpellNames["Blood Plague"],
				["BF"] = SpellNames["Blood Plague"]..", "..SpellNames["Frost Fever"],
				["F"] = SpellNames["Frost Fever"],
				["B"] = SpellNames["Blood Plague"],
			},
			set = function(info, val)
				self.db.profile.diseaseOrder = val
				self.displays.target:UpdateLayout()
			end,
			get = function(info)
				return self.db.profile.diseaseOrder
			end,
		},
		specHdr = {
			order = 200,
			type = "header",
			name = L["Enabled for Specs"],
		},
		specBlood = {
			name = L["Blood"],
			desc = L["Blood"],
			type = "toggle",
			order = 210,
			set = function(info, val)
				self.db.profile.specs["Blood"] = val
				self:CheckEnabled()
			end,
			get = function(info)
				return self.db.profile.specs["Blood"]
			end,
		},
		specFrost = {
			name = L["Frost"],
			desc = L["Frost"],
			type = "toggle",
			order = 220,
			set = function(info, val)
				self.db.profile.specs["Frost"] = val
				self:CheckEnabled()
			end,
			get = function(info)
				return self.db.profile.specs["Frost"]
			end,
		},
		specUnholy = {
			name = L["Unholy"],
			desc = L["Unholy"],
			type = "toggle",
			order = 230,
			set = function(info, val)
				self.db.profile.specs["Unholy"] = val
				self:CheckEnabled()
			end,
			get = function(info)
				return self.db.profile.specs["Unholy"]
			end,
		},
		--ldb = {
		--	order = 300,
		--	type = "header",
		--	name = L["LDB"],
		--},
		--LDBDataFeed = {
		--	order = 310,
		--	name = L["Data Feed"],
		--	desc = L["DataFeed_OptionDesc"],
		--	type = "select",
		--	values = {
		--	    ["None"] = L["None"],
		--	},
		--	set = function(info, val)
		--		self.db.profile.LDBDataDisplay = val
		--		self:SetLDBDataDisplay()
		--	    UpdateLDBData()
		--	end,
		--	get = function(info)
		--		return self.db.profile.LDBDataDisplay
		--	end,
		--},
	}
	return options
end

function DKDiseases:OnInitialize()
	self:CheckClass()
	if not addon.isDK then return end

	-- Load the settings
	self.db = _G.LibStub("AceDB-3.0"):New("DKDiseasesDB", defaults, "Default")
	addon.db = self.db
	self:SetLDBDataDisplay()

	--self:MigrateSettings()

	-- Register for profile callbacks
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChange")
	
	Icon:Register("DKDiseasesLDB", Broker.obj, self.db.profile.minimap)
	LSM.RegisterCallback(DKDiseases, "LibSharedMedia_Registered")

	for name, obj in pairs(addon.modules) do
		if obj and obj.OnInitialize then
			obj:OnInitialize()
		end
	end
end

function DKDiseases:OnProfileChange()
	--self:MigrateSettings()
	self:Reset()
	addon:FireCallback("ProfileUpdate")
end

function DKDiseases:Reset()
	--self:ResetFonts()
	--self:UpdateTextures()

	for name, display in pairs(self.registeredDisplays) do
		if display and display.Reset then
			display:Reset()
		end
	end
end

function DKDiseases:LibSharedMedia_Registered(event, mediatype, key)
	if _G.strlen(self.db.profile.font_face) > 1 and mediatype == "font" then
		if self.db.profile.font_face == key then
			self:ResetFonts()
		end
	end
	if mediatype == "statusbar" then
	    self:UpdateTextures()
	end
end

function DKDiseases:ResetFonts()
	for name, display in pairs(self.registeredDisplays) do
		if display and display.ResetFonts then
			display:ResetFonts()
		end
	end
end

function DKDiseases:UpdateTextures()
	for name, display in pairs(self.registeredDisplays) do
		if display and display.UpdateTexture then
			display:UpdateTexture()
		end
	end
end

local function splitWords(str)
  local w = {}
  local function helper(word) table.insert(w, word) return nil end
  str:gsub("(%w+)", helper)
  return w
end

function DKDiseases:ChatCommand(input)
	if not input or input:trim() == "" then
		self:ShowOptions()
	else
		local cmds = splitWords(input)
		if cmds[1] and cmds[1] == "debug" then
			if cmds[2] and cmds[2] == "on" then
				self.db.profile.debug = true
				self:Print("Debugging on.  Use '/dkdiseases debug off' to disable.")
			elseif cmds[2] and cmds[2] == "off" then
				self.db.profile.debug = false
				self:Print("Debugging off.")
			else
				self:Print("Debugging is "..(self.db.profile.debug and "on." or "off."))
			end
		elseif cmds[1] and cmds[1] == "srvshb" then
			local fmt = "SR vs HB: SR = %s [%s + %s], HB = %s"
			self:Print(fmt:format(
				FormatNumber((addon.soulReaper1 or 0)+(addon.soulReaper2 or 0)),
				FormatNumber(addon.soulReaper1 or 0), 
				FormatNumber(addon.soulReaper2 or 0), 
				FormatNumber(addon.howlingBlast or 0)))
		end
	end
end

local MandatoryEvents = {
	"PLAYER_TALENT_UPDATE",
	"PLAYER_SPECIALIZATION_CHANGED",
	"UNIT_LEVEL",
	"UNIT_ENTERED_VEHICLE",
	"UNIT_EXITED_VEHICLE",
	"PLAYER_GAINS_VEHICLE_DATA",
	"PLAYER_LOSES_VEHICLE_DATA",
}
function DKDiseases:OnEnable()
	if not addon.isDK then return end

	for i = 1, #MandatoryEvents do
		self:RegisterEvent(MandatoryEvents[i])
	end
	LoadSpellNames()
	self:CheckTalents()
	--self:ToggleStatTracking()
	--self:CheckEnabled()

	if not self.optionsFrame then
		-- Register Options
		local displayName = addon.addonTitle
		self:GetOptions()
		_G.LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable(
		displayName, addon.options)

		self.optionsFrame = {}
		local ACD = _G.LibStub("AceConfigDialog-3.0")
		self.optionsFrame.Main = ACD:AddToBlizOptions(
			displayName, displayName, nil, "core")
		--self.optionsFrame.Layout = ACD:AddToBlizOptions(
		--    displayName, L["Layout"], displayName, "layout")

		ACD:AddToBlizOptions(
			displayName, addon.options.args.profile.name, displayName, "profile")
		-- Register the chat command
		self:RegisterChatCommand("dkdiseases", "ChatCommand")
	end

	for name, obj in pairs(addon.modules) do
		if obj and obj.Enable then
			obj:Enable()
		end
	end
end

local WatchedEvents = {
	"PLAYER_REGEN_DISABLED",
	"PLAYER_REGEN_ENABLED",
	"PLAYER_TARGET_CHANGED",
	"PLAYER_FOCUS_CHANGED",
	"PLAYER_ENTERING_WORLD",
}
local WatchedUnitEvents = {
	["target"] = {"UNIT_AURA"},
	["focus"] = {"UNIT_AURA"},
}
local function EventFrame_OnEvent(frame, event, ...)
	if event == "UNIT_ATTACK_POWER" then
		DKDiseases:UNIT_ATTACK_POWER(event, ...)
	elseif event == "UNIT_STATS" then
		DKDiseases:UNIT_STATS(event, ...)
	elseif event == "UNIT_AURA" then
		DKDiseases:UNIT_AURA(event, ...)
	end
end
local EventFrames = {}
local WatchedUnits = { "player", "target", "focus" }
for i, unit in ipairs(WatchedUnits) do
	if not EventFrames[unit] then
		local frame = _G.CreateFrame("Frame", ADDON_NAME.."_EventFrame_"..unit)
		frame:SetScript("OnEvent", EventFrame_OnEvent)
		EventFrames[unit] = frame
	end
end

function DKDiseases:Load()
	if addon.isDK then
		self:CreateOrUpdateAllDisplays()
		if self.db.profile.specs[addon.currentSpec] then
			for name, display in pairs(self.displays) do
				if display and display.Show then
					display:Show()
				end
			end
			for i = 1, #WatchedEvents do
				self:RegisterEvent(WatchedEvents[i])
			end
			for unit, events in pairs(WatchedUnitEvents) do
				local eventFrame = EventFrames[unit]
				if eventFrame then
					for i, event in ipairs(events) do
						eventFrame:RegisterUnitEvent(event, unit)
					end
				else
					self:Print("Missing event frame for "..tostring(unit).."!")
				end
			end
		end
		addon:FireCallback("Load")
	end
end

function DKDiseases:Unload()
	for i = 1, #WatchedEvents do
		self:UnregisterEvent(WatchedEvents[i])
	end
	for unit, events in pairs(WatchedUnitEvents) do
		local eventFrame = EventFrames[unit]
		if eventFrame then
			for i, event in ipairs(events) do
				eventFrame:UnregisterEvent(event)
			end
		end
	end
	for name, display in pairs(self.displays) do
		if display and display.Disable then display:Disable() end
	end
end

local StatUnitEvents = {
	["player"] = {"UNIT_ATTACK_POWER", "UNIT_STATS", "UNIT_AURA"},
}
local StatEvents = {
	"PLAYER_DAMAGE_DONE_MODS",
	"COMBAT_RATING_UPDATE",
}
function DKDiseases:ToggleStatTracking(flag)
	local enable = flag ~= nil and flag or addon.db.profile.srFrame.enabled
	if addon.isDK and enable then
		for i = 1, #StatEvents do
			self:RegisterEvent(StatEvents[i])
		end
		for unit, events in pairs(StatUnitEvents) do
			local eventFrame = EventFrames[unit]
			if eventFrame then
				for i, event in ipairs(events) do
					eventFrame:RegisterUnitEvent(event, unit)
				end
			else
				self:Print("Missing event frame for "..tostring(unit).."!")
			end
		end
		self:UpdateAllStats()
	else
		for i = 1, #StatEvents do
			self:UnregisterEvent(StatEvents[i])
		end
		for unit, events in pairs(StatUnitEvents) do
			local eventFrame = EventFrames[unit]
			if eventFrame then
				for i, event in ipairs(events) do
					eventFrame:UnregisterEvent(event)
				end
			end
		end
	end
end

function DKDiseases:OnDisable()
	self:Unload()
end

function DKDiseases:CheckEnabled()
	if addon.isDK and self.db.profile.specs[addon.currentSpec] then
		self:Load()
	else
		self:Unload()
	end
end

function DKDiseases:CheckClass()
	addon.isDK = false
	local class, className = _G.UnitClass("player")
	if className then
		if (className == 'DEATH KNIGHT' or className == 'DEATHKNIGHT') then
			addon.isDK = true
		end
	end
end

function DKDiseases:CheckTalents()
	self:CheckClass()
	if addon.isDK then
		local activeSpecNum = _G.GetSpecialization()
		if activeSpecNum and activeSpecNum > 0 then
			local id, name, desc, texture = _G.GetSpecializationInfo(activeSpecNum)
			if texture == "Interface\\Icons\\Spell_Deathknight_BloodPresence" then
				addon.currentSpec = "Blood"
			elseif texture == "Interface\\Icons\\Spell_Deathknight_FrostPresence" then
				addon.currentSpec = "Frost"
			elseif texture == "Interface\\Icons\\Spell_Deathknight_UnholyPresence" then
				addon.currentSpec = "Unholy"
			else
				self:Print("Error detecing player spec.")
				addon.currentSpec = "Blood"
			end
		end

		addon.necroticPlague = false
		local activeSpecGroup = _G.GetActiveSpecGroup()
		if activeSpecGroup then
			addon.necroticPlague = addon.HasActiveTalent("Necrotic Plague")
		end
	end
	
	self:CheckEnabled()

	addon:FireCallback("TalentUpdate")

	self:UpdateAllDisplays()

	if self.db.profile.debug then
		local fmt = "CheckTalents [isDK=%s,spec=%s]"
		self:Print(fmt:format(tostring(addon.isDK), addon.currentSpec))
	end
end

local fontFlags = {}
function DKDiseases:GetFontSettings()
	wipe(fontFlags)
	for k, v in pairs(self.db.profile.fontFlags) do
		if v then tinsert(fontFlags, k) end
	end
	return self.db.profile.font_face, self.db.profile.font_size, 
		tconcat(fontFlags, ",")
end

function DKDiseases:RegisterDisplay(name, object)
	self.registeredDisplays[name] = object
end

function DKDiseases:UpdateAllDisplays()
	for i, unit in ipairs(addon.units) do
		if self.db.profile.units[unit].enabled then
			if self.displays[unit] then
				self.displays[unit]:UpdateLayout()
			end
		end
	end
end

function DKDiseases:CreateOrUpdateAllDisplays()
	for i, unit in ipairs(addon.units) do
		self:CreateOrUpdateDisplay(unit)
	end
end

function DKDiseases:CreateOrUpdateDisplay(unit)
	if not unit then return end

	if not self.db.profile.units[unit].enabled then
		if self.displays[unit] then
			self.displays[unit]:Disable()
		end
		return
	end

	local prev = nil
	if self.displays[unit] then
		prev = self.displays[unit].name
	end

	local display = self.db.profile.display[unit]
	if display and display ~= prev then
		if self.displays[unit] then
			self.displays[unit]:Disable()
		end

		local settings = self.db.profile.displays[display][unit]
		local registered = self.registeredDisplays[display]
		if settings and registered then
			local obj = registered:Create(unit)
			if obj then
				self.displays[unit] = obj
				self.diseases[unit].ff = obj.ff
				self.diseases[unit].bp = obj.bp
				self.diseases[unit].np = obj.np
				obj:Enable()
			end
			if not self.displays[unit] then
				self:Print("Could not create the display!")
			end
		else
			self:Print("Could not find registered display!")
		end
	elseif display and display == prev then
		if self.displays[unit] and self.displays[unit].Show then
			self.displays[unit]:Show()
		end
	end
end

function DKDiseases:PLAYER_DEAD()
	addon:FireCallback("PlayerDead")
end

function DKDiseases:PLAYER_ALIVE()
	addon:FireCallback("PlayerAlive")
end

function DKDiseases:UNIT_LEVEL(event, unit)
	if unit == "player" then
		self:CheckTalents()
	end
end

function DKDiseases:UNIT_ATTACK_POWER(event, unit)
	if unit == "player" then
		self:UpdateAllStats()
	end
end

function DKDiseases:UNIT_STATS(event, unit)
	if unit == "player" then
		self:UpdateAllStats()
	end
end

function DKDiseases:PLAYER_DAMAGE_DONE_MODS(event, unit)
	if unit == "player" then
		self:UpdateAllStats()
	end
end

function DKDiseases:COMBAT_RATING_UPDATE()
	self:UpdateAllStats()
end

function DKDiseases:UpdateAllStats()
	local update = false

	local base, posBuff, negBuff = _G.UnitAttackPower("player")
	local effAP = base + posBuff - negBuff
	if effAP ~= addon.effectiveAP then
		addon.effectiveAP = effAP
		update = true
	end

	local wpnSpeed = _G.select(1, _G.UnitAttackSpeed("player")) * 
		(1 + _G.GetMeleeHaste() / 100)
	local apAdj = effAP / 14 * wpnSpeed
	local min, max = _G.UnitDamage("player")
	local avgWeaponDmg = (min + max) / 2 - apAdj
	if avgWeaponDmg ~= addon.avgWeaponDmg then
		addon.avgWeaponDmg = avgWeaponDmg
		update = true
	end

	local dmgMod = _G.select(7, _G.UnitDamage("player")) or 1.0
	if dmgMod ~= addon.damageModifier then
		addon.damageModifier = dmgMod
		update = true
	end

	local mastery = _G.GetMastery()
	if mastery ~= addon.mastery then
		addon.mastery = mastery
		update = true
	end

	local critChance = _G.GetCritChance() / 100
	if critChance > 1 then critChance = 1 end
	if critChance ~= addon.critChance then
		addon.critChance = critChance
		update = true
	end

	-- Get the spell crit chance for the Frost school
	local spellCrit = _G.GetSpellCritChance(5) / 100
	if spellCrit > 1 then spellCrit = 1 end
	if spellCrit ~= addon.spellCrit then
		addon.spellCrit = spellCrit
		update = true
	end

	if update then addon:FireCallback("StatUpdate") end
end

function DKDiseases:PLAYER_REGEN_DISABLED()
	-- Player has entered combat.
	for unit, display in pairs(self.displays) do
		if self.displays[unit] then
			self.displays[unit]:OnCombatEnter()
		end
	end
	addon:FireCallback("CombatStart")
end

function DKDiseases:PLAYER_REGEN_ENABLED()
	-- Player has left combat.
	for unit, display in pairs(self.displays) do
		if self.displays[unit] then
			self.displays[unit]:OnCombatLeave()
		end
	end
	addon:FireCallback("CombatEnd")
end

function DKDiseases:PLAYER_TARGET_CHANGED(unit)
	addon.current.target = _G.UnitGUID("target")
	addon.hostile.target = _G.UnitExists("target") and 
		_G.UnitCanAttack("player", "target") and
		not _G.UnitIsDead("target")
	if self.displays.target and self.displays.target.Show then
		self.displays.target:Show()
	end
	if self.displays.target and self.db.profile.units.target.enabled then
		if not addon.current.target and self.displays.target.ClearDisplay then
			self.displays.target:ClearDisplay()
		end
		self:UNIT_AURA("UNIT_AURA", "target")
	end
	addon:FireCallback("TargetChanged")
end

function DKDiseases:PLAYER_FOCUS_CHANGED(unit)
	addon.current.focus = _G.UnitGUID("focus")
	addon.hostile.focus = _G.UnitExists("target") and 
		_G.UnitCanAttack("player", "target") and
		not _G.UnitIsDead("target")
	if self.displays.focus and self.displays.focus.Show then 
		self.displays.focus:Show()
	end
	if self.displays.focus and self.db.profile.units.focus.enabled then
		if not addon.current.focus and self.displays.target.ClearDisplay then
			self.displays.focus:ClearDisplay()
		end
		self:UNIT_AURA("UNIT_AURA", "focus")
	end
end

local stdDiseases, npDiseases
local magicDebuffs
function DKDiseases:UNIT_AURA(event, unit)
	local name, rank, icon, count, dispelType, duration, expires, 
		caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
		isBossDebuff, castByPlayer, value1, value2, value3 

	if unit == "target" or unit == "focus" then
		stdDiseases = stdDiseases or {
			[SpellNames["Frost Fever"]] = "ff",
			[SpellNames["Blood Plague"]] = "bp",
		}
		npDiseases = npDiseases or {
			[SpellNames["Necrotic Plague"]] = "np",
		}
		local diseases = addon.necroticPlague and npDiseases or stdDiseases

		for spell, id in pairs(diseases) do
			local display = self.diseases[unit][id]
			if display then
				name, rank, icon, count, dispelType, duration, expires, 
				caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
				isBossDebuff, castByPlayer, value1, value2, value3 
				= _G.UnitDebuff(unit, spell, nil, "PLAYER")

				if name then
					local timeLeft = expires - _G.GetTime()
					display:OnDiseaseRefreshed(duration, expires, count)
				else
					display:OnDiseaseRemoved()
				end
			end
		end

		if unit == "target" then
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
			= _G.UnitDebuff(unit, SpellNames["Frost Vulnerability"], nil, "PLAYER")
			if name then
				addon.frostVulnerability = 1 + (0.03 * (count or 0))
			else
				addon.frostVulnerability = 1.0
			end
		end
	end
end

function DKDiseases:PLAYER_SPECIALIZATION_CHANGED(event)
	self:CheckTalents()
end

function DKDiseases:PLAYER_TALENT_UPDATE(event)
	self:CheckTalents()
end

function DKDiseases:PLAYER_ENTERING_WORLD(event)
	addon.playerGUID = _G.UnitGUID("player") or ""
	self:UNIT_AURA("PLAYER_ENTERING_WORLD", "player")
end

function DKDiseases:CheckPlayerStatus()
	local changed = false
	local inVehicle = _G.UnitUsingVehicle("player")
	if inVehicle ~= addon.inVehicle then
		addon.inVehicle = inVehicle
		changed = true
	end
	if changed then
		for unit, display in pairs(self.displays) do
			if display and display.Show then
				display:Show()
			end
		end
	end
end

function DKDiseases:UNIT_ENTERED_VEHICLE(event)
	self:CheckPlayerStatus()
end

function DKDiseases:UNIT_EXITED_VEHICLE(event)
	self:CheckPlayerStatus()
end

function DKDiseases:PLAYER_GAINS_VEHICLE_DATA(event)
	self:CheckPlayerStatus()
end

function DKDiseases:PLAYER_LOSES_VEHICLE_DATA(event)
	self:CheckPlayerStatus()
end
